# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["multivariate_gaussian_samples", "nd_sort_samples"]

import numpy as np
from scipy.spatial import cKDTree


def multivariate_gaussian_samples(matrix, N, mean=None):
    """
    Generate samples from a multidimensional Gaussian with a given covariance.

    :param matrix: ``(k, k)``
        The covariance matrix.

    :param N:
        The number of samples to generate.

    :param mean: ``(k,)`` (optional)
        The mean of the Gaussian. Assumed to be zero if not given.

    :returns samples: ``(k,)`` or ``(N, k)``
        Samples from the given multivariate normal.

    """
    if mean is None:
        mean = np.zeros(len(matrix))
    samples = np.random.multivariate_normal(mean, matrix, N)
    if N == 1:
        return samples[0]
    return samples


def nd_sort_samples(samples):
    """
    Sort an N-dimensional list of samples using a KDTree.

    :param samples: ``(nsamples, ndim)``
        The list of samples. This must be a two-dimensional array.

    :returns i: ``(nsamples,)``
        The list of indices into the original array that return the correctly
        sorted version.

    """
    # Check the shape of the sample list.
    assert len(samples.shape) == 2

    # Build a KD-tree on the samples.
    tree = cKDTree(samples)

    # Compute the distances.
    d, i = tree.query(samples[0], k=len(samples))
    return i


def numerical_gradient(f, x, dx=1.234e-6):
    g = np.empty_like(x, dtype=float)
    for i in range(len(g)):
        x[i] += dx
        fp = f(x)
        x[i] -= 2*dx
        fm = f(x)
        x[i] += dx
        g[i] = 0.5 * (fp - fm) / dx
    return g


def check_gradient(obj, *args, **kwargs):
    eps = kwargs.pop("eps", 1.23e-5)

    grad0 = obj.get_gradient(*args, **kwargs)
    vector = obj.get_parameter_vector()
    for i, v in enumerate(vector):
        # Compute the centered finite difference approximation to the gradient.
        vector[i] = v + eps
        obj.set_parameter_vector(vector)
        p = obj.get_value(*args, **kwargs)

        vector[i] = v - eps
        obj.set_parameter_vector(vector)
        m = obj.get_value(*args, **kwargs)

        vector[i] = v
        obj.set_parameter_vector(vector)

        grad = 0.5 * (p - m) / eps
        assert np.allclose(grad0[i], grad), \
            "grad computation failed for '{0}' ({1})" \
            .format(obj.get_parameter_names()[i], i)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = [
    "Kernel", "Sum", "Product",
    {%- for spec in specs %}
    "{{ spec.name }}",
    {%- endfor %}
]

import numpy as np

from .modeling import Model, ModelSet
from .metrics import Metric, Subspace
from .kernel_interface import KernelInterface


class Kernel(ModelSet):
    """
    The abstract kernel type. Every kernel implemented in George should be
    a subclass of this object.

    """

    is_kernel = True
    kernel_type = -1

    # This function deals with weird behavior when performing arithmetic
    # operations with numpy scalars.
    def __array_wrap__(self, array, context=None):
        if context is None:
            raise TypeError("Invalid operation")
        ufunc, args, _ = context
        if ufunc.__name__ == "multiply":
            return float(args[0]) * args[1]
        elif ufunc.__name__ == "add":
            return float(args[0]) + args[1]
        raise TypeError("Invalid operation")
    __array_priority__ = np.inf

    def __getstate__(self):
        odict = self.__dict__.copy()
        odict["_kernel"] = None
        return odict

    # We must overload the ModelSet attribute getter to pass the requests
    # to the "BaseKernel"
    def __getattr__(self, name):
        if "models" in self.__dict__:
            if name in self.models:
                return self.models[name]
            if None in self.models:
                return getattr(self.models[None], name)
        raise AttributeError(name)

    @property
    def kernel(self):
        return KernelInterface(self)

    def __repr__(self):
        kernel = self.models[None]
        params = ["{0}={1}".format(k, getattr(kernel, k))
                  for k in kernel.parameter_names]
        if self.stationary:
            params += ["metric={0}".format(repr(self.metric)),
                       "block={0}".format(repr(self.block))]
        else:
            params += ["ndim={0}".format(self.ndim),
                       "axes={0}".format(repr(self.axes))]
        return "{0}({1})".format(self.__class__.__name__, ", ".join(params))

    def __add__(self, b):
        if not hasattr(b, "is_kernel"):
            return Sum(ConstantKernel(log_constant=np.log(float(b)/self.ndim),
                                      ndim=self.ndim), self)
        return Sum(self, b)

    def __radd__(self, b):
        return self.__add__(b)

    def __mul__(self, b):
        if not hasattr(b, "is_kernel"):
            log_constant = np.log(float(b)/self.ndim)
            return Product(ConstantKernel(log_constant=log_constant,
                                          ndim=self.ndim), self)
        return Product(self, b)

    def __rmul__(self, b):
        return self.__mul__(b)

    def get_value(self, x1, x2=None, diag=False):
        x1 = np.ascontiguousarray(x1, dtype=np.float64)
        if x2 is None:
            if diag:
                return self.kernel.value_diagonal(x1, x1)
            else:
                return self.kernel.value_symmetric(x1)
        x2 = np.ascontiguousarray(x2, dtype=np.float64)
        if diag:
            return self.kernel.value_diagonal(x1, x2)
        else:
            return self.kernel.value_general(x1, x2)

    def get_gradient(self, x1, x2=None, include_frozen=False):
        mask = (
            np.ones(self.full_size, dtype=bool)
            if include_frozen else self.unfrozen_mask
        )
        which = mask.astype(np.uint32)
        x1 = np.ascontiguousarray(x1, dtype=np.float64)
        if x2 is None:
            g = self.kernel.gradient_symmetric(which, x1)
        else:
            x2 = np.ascontiguousarray(x2, dtype=np.float64)
            g = self.kernel.gradient_general(which, x1, x2)
        return g[:, :, mask]

    def get_x1_gradient(self, x1, x2=None):
        x1 = np.ascontiguousarray(x1, dtype=np.float64)
        if x2 is None:
            x2 = x1
        else:
            x2 = np.ascontiguousarray(x2, dtype=np.float64)
        return self.kernel.x1_gradient_general(x1, x2)

    def get_x2_gradient(self, x1, x2=None):
        x1 = np.ascontiguousarray(x1, dtype=np.float64)
        if x2 is None:
            x2 = x1
        else:
            x2 = np.ascontiguousarray(x2, dtype=np.float64)
        return self.kernel.x2_gradient_general(x1, x2)

    def test_gradient(self, x1, x2=None, eps=1.32e-6, **kwargs):
        vector = self.get_parameter_vector()
        g0 = self.get_gradient(x1, x2=x2)

        for i, v in enumerate(vector):
            vector[i] = v + eps
            self.set_parameter_vector(vector)
            kp = self.get_value(x1, x2=x2)

            vector[i] = v - eps
            self.set_parameter_vector(vector)
            km = self.get_value(x1, x2=x2)

            vector[i] = v
            self.set_parameter_vector(vector)

            grad = 0.5 * (kp - km) / eps
            assert np.allclose(g0[:, :, i], grad, **kwargs), \
                "incorrect gradient for parameter '{0}' ({1})" \
                .format(self.get_parameter_names()[i], i)

    def test_x1_gradient(self, x1, x2=None, eps=1.32e-6, **kwargs):
        kwargs["atol"] = kwargs.get("atol", 0.5 * eps)
        g0 = self.get_x1_gradient(x1, x2=x2)
        if x2 is None:
            x2 = np.array(x1)
        for i in range(len(x1)):
            for k in range(self.ndim):
                x1[i, k] += eps
                kp = self.get_value(x1, x2=x2)

                x1[i, k] -= 2*eps
                km = self.get_value(x1, x2=x2)

                x1[i, k] += eps

                grad = 0.5 * (kp - km) / eps
                assert np.allclose(g0[i, :, k], grad[i], **kwargs)

    def test_x2_gradient(self, x1, x2=None, eps=1.32e-6, **kwargs):
        kwargs["atol"] = kwargs.get("atol", 0.5 * eps)
        g0 = self.get_x2_gradient(x1, x2=x2)
        if x2 is None:
            x2 = np.array(x1)
        for i in range(len(x2)):
            for k in range(self.ndim):
                x2[i, k] += eps
                kp = self.get_value(x1, x2=x2)

                x2[i, k] -= 2*eps
                km = self.get_value(x1, x2=x2)

                x2[i, k] += eps

                grad = 0.5 * (kp - km) / eps
                assert np.allclose(g0[:, i, k], grad[:, i], **kwargs)


class _operator(Kernel):
    is_kernel = False
    kernel_type = -1
    operator_type = -1

    def __init__(self, k1, k2):
        if k1.ndim != k2.ndim:
            raise ValueError("Dimension mismatch")
        self.ndim = k1.ndim
        self._dirty = True
        super(_operator, self).__init__([("k1", k1), ("k2", k2)])

    @property
    def k1(self):
        return self.models["k1"]

    @property
    def k2(self):
        return self.models["k2"]

    @property
    def dirty(self):
        return self._dirty or self.k1.dirty or self.k2.dirty

    @dirty.setter
    def dirty(self, v):
        self._dirty = v
        self.k1.dirty = False
        self.k2.dirty = False


class Sum(_operator):
    is_kernel = False
    operator_type = 0

    def __repr__(self):
        return "{0} + {1}".format(self.k1, self.k2)


class Product(_operator):
    is_kernel = False
    operator_type = 1

    def __repr__(self):
        return "{0} * {1}".format(self.k1, self.k2)

{% for spec in specs %}
class Base{{ spec.name }} (Model):
    parameter_names = ({% for p in spec.params -%}"{{ p }}", {% endfor %})


class {{ spec.name }} (Kernel):
    r"""
    {{ spec.doc | indent(4) }}

    """

    kernel_type = {{ spec.index }}
    stationary = {{ spec.stationary }}

    def __init__(self,
                 {% for p in spec.params %}{{ p }}=None,
                 {% endfor -%}
                 {% for con in spec.constants %}{{ con.name }}=None,
                 {% endfor -%}
                 {% if spec.stationary -%}
                 metric=None,
                 metric_bounds=None,
                 lower=True,
                 block=None,
                 {% endif -%}
                 bounds=None,
                 ndim=1,
                 axes=None):
        {% for con in spec.constants %}
        if {{ con.name }} is None:
            raise ValueError("missing required parameter '{{ con.name }}'")
        self.{{ con.name }} = {{ con.name }}
        {% endfor %}
        {% if spec.stationary -%}
        if metric is None:
            raise ValueError("missing required parameter 'metric'")
        metric = Metric(metric, bounds=metric_bounds, ndim=ndim,
                        axes=axes, lower=lower)
        self.ndim = metric.ndim
        self.axes = metric.axes
        self.block = block
        {%- else -%}
        self.subspace = Subspace(ndim, axes=axes)
        self.ndim = self.subspace.ndim
        self.axes = self.subspace.axes
        {%- endif %}

        kwargs = dict({% for p in spec.params -%}{{ p }}={{ p }}, {% endfor -%})
        if bounds is not None:
            kwargs["bounds"] = bounds
        base = Base{{ spec.name }}(**kwargs)
        super({{ spec.name }}, self).__init__([
            (None, base), {% if spec.stationary -%}("metric", metric){%- endif %}
        ])

        # Common setup.
        self.dirty = True
    {% if spec.stationary %}
    @property
    def block(self):
        if not self.blocked:
            return None
        return list(zip(self.min_block, self.max_block))

    @block.setter
    def block(self, block):
        if block is None:
            self.blocked = False
            self.min_block = -np.inf + np.zeros(len(self.axes))
            self.max_block = np.inf + np.zeros(len(self.axes))
            return

        block = np.atleast_2d(block)
        if block.shape != (len(self.axes), 2):
            raise ValueError("dimension mismatch in block specification")
        self.blocked = True
        self.min_block, self.max_block = map(np.array, zip(*block))
    {% endif %}
{% endfor %}



# -*- coding: utf-8 -*-

import os
import glob
import yaml
import george

# Inject the kernel docs
d = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
with open(os.path.join(d, "docs", "user", "kernels.rst.template"), "r") as f:
    TEMPLATE = f.read()

fns = glob.glob(os.path.join(d, "kernels", "*.yml"))
if len(fns):
    specs = []
    for i, fn in enumerate(fns):
        with open(fn, "r") as f:
            specs.append(yaml.safe_load(f.read()))
    tokens = []
    for spec in specs:
        if spec["stationary"]:
            tokens += [
                ".. autoclass:: george.kernels.{0}".format(spec["name"])
            ]
    TEMPLATE = TEMPLATE.replace("STATIONARYKERNELS", "\n".join(tokens))
    tokens = []
    for spec in specs:
        if not spec["stationary"]:
            tokens += [
                ".. autoclass:: george.kernels.{0}".format(spec["name"])
            ]
    TEMPLATE = TEMPLATE.replace("OTHERKERNELS", "\n".join(tokens))

with open(os.path.join(d, "docs", "user", "kernels.rst"), "w") as f:
    f.write(TEMPLATE)

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
]
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"

# General information about the project.
project = "george"
copyright = "2012-2023 Dan Foreman-Mackey"

version = george.__version__
release = george.__version__

exclude_patterns = ["_build"]
pygments_style = "sphinx"

# Readthedocs.
html_theme = "pydata_sphinx_theme"
html_title = "george"
htmp_theme_options = dict(
    analytics_id="analytics_id",
)
# html_context = dict(
#     display_github=True,
#     github_user="dfm",
#     github_repo="george",
#     github_version="main",
#     conf_py_path="/docs/",
# )
html_static_path = ["_static"]
# html_show_sourcelink = False



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["HODLRSolver"]

import numpy as np

from .basic import BasicSolver
from ._hodlr import HODLRSolver as HODLRSolverInterface


class HODLRSolver(BasicSolver):
    r"""
    A solver using `Sivaram Amambikasaran's HODLR algorithm
    <http://arxiv.org/abs/1403.6015>`_ to approximately solve the GP linear
    algebra in :math:`\mathcal{O}(N\,\log^2 N)`.

    :param kernel:
        An instance of a subclass of :class:`kernels.Kernel`.
    :param min_size: (optional[int])
        The block size where the solver switches to a general direct 
        factorization algorithm. This can be tuned for platform and 
        problem specific performance and accuracy. As a general rule,
        larger values will be more accurate and slower, but there is some
        overhead for very small values, so we recommend choosing values in the
        hundreds. (default: ``100``)
    :param tol: (optional[float])
        The precision tolerance for the low-rank approximation. 
        This value is used as an approximate limit on the Frobenius norm 
        between the low-rank approximation and the true matrix
        when reconstructing the off-diagonal blocks. Smaller values of ``tol``
        will generally give more accurate results with higher computational
        cost. (default: ``0.1``)
    :param seed: (optional[int])
        The low-rank approximation method within the HODLR algorithm
        is not deterministic and, without a fixed seed, the method
        can give different results for the same matrix. Therefore, we require
        that the user provide a seed for the random number generator.
        (default: ``42``)
    """

    def __init__(self, kernel, min_size=100, tol=0.1, seed=42):
        self.min_size = min_size
        self.tol = tol
        self.seed = seed
        super(HODLRSolver, self).__init__(kernel)

    def compute(self, x, yerr):
        self.solver = HODLRSolverInterface()
        self.solver.compute(self.kernel, x, yerr,
                            self.min_size, self.tol, self.seed)
        self._log_det = self.solver.log_determinant
        self.computed = self.solver.computed

    def apply_inverse(self, y, in_place=False):
        return self.solver.apply_inverse(y, in_place=in_place)

    def dot_solve(self, y):
        return self.solver.dot_solve(y)

    def apply_sqrt(self, r):
        raise NotImplementedError("apply_sqrt is not implemented for the "
                                  "HODLRSolver")

    def get_inverse(self):
        return self.solver.get_inverse()

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_computed"] = False
        del state["solver"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)



# -*- coding: utf-8 -*-

from __future__ import division, print_function, absolute_import

__all__ = ["test_tutorial"]

import numpy as np

from george import kernels, GP, BasicSolver, HODLRSolver


def test_tutorial():
    def model(params, t):
        _, _, amp, loc, sig2 = params
        return amp * np.exp(-0.5 * (t - loc) ** 2 / sig2)

    def lnlike(p, t, y, yerr, solver=BasicSolver):
        a, tau = np.exp(p[:2])
        gp = GP(a * kernels.Matern32Kernel(tau) + 0.001, solver=solver)
        gp.compute(t, yerr)
        return gp.lnlikelihood(y - model(p, t))

    def lnprior(p):
        lna, lntau, amp, loc, sig2 = p
        if (-5 < lna < 5 and -5 < lntau < 5 and -10 < amp < 10 and
                -5 < loc < 5 and 0 < sig2 < 3):
            return 0.0
        return -np.inf

    def lnprob(p, x, y, yerr, **kwargs):
        lp = lnprior(p)
        return lp + lnlike(p, x, y, yerr, **kwargs) \
            if np.isfinite(lp) else -np.inf

    np.random.seed(1234)
    x = np.sort(np.random.rand(50))
    yerr = 0.05 + 0.01 * np.random.rand(len(x))
    y = np.sin(x) + yerr * np.random.randn(len(x))
    p = [0, 0, -1.0, 0.1, 0.4]
    assert np.isfinite(lnprob(p, x, y, yerr)), "Incorrect result"
    assert np.allclose(lnprob(p, x, y, yerr),
                       lnprob(p, x, y, yerr, solver=HODLRSolver)), \
        "Inconsistent results"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import glob
import yaml
from jinja2 import Template


def compile_kernels(fns):
    template_dir = "templates"
    output_dir = os.path.join("src", "george")

    with open(os.path.join(template_dir, "parser.h")) as f:
        PARSER_TEMPLATE = Template(f.read())
    with open(os.path.join(template_dir, "kernels.h")) as f:
        CPP_TEMPLATE = Template(f.read())
    with open(os.path.join(template_dir, "kernels.py")) as f:
        PYTHON_TEMPLATE = Template(f.read())

    specs = []
    for i, fn in enumerate(fns):
        with open(fn, "r") as f:
            spec = yaml.load(f.read(), Loader=yaml.FullLoader)
        print("Found kernel '{0}'".format(spec["name"]))
        spec["index"] = i
        spec["reparams"] = spec.get("reparams", {})
        specs.append(spec)
    print("Found {0} kernel specifications".format(len(specs)))

    fn = os.path.join(output_dir, "include", "george", "parser.h")
    with open(fn, "w") as f:
        print("Saving parser to '{0}'".format(fn))
        f.write(PARSER_TEMPLATE.render(specs=specs))
    fn = os.path.join(output_dir, "include", "george", "kernels.h")
    with open(fn, "w") as f:
        print("Saving C++ kernels to '{0}'".format(fn))
        f.write(CPP_TEMPLATE.render(specs=specs))
    fn = os.path.join(output_dir, "kernels.py")
    with open(fn, "w") as f:
        print("Saving Python kernels to '{0}'".format(fn))
        f.write(PYTHON_TEMPLATE.render(specs=specs))


if __name__ == "__main__":
    # If the kernel specifications are included (development mode) re-compile
    # them first.
    kernel_specs = glob.glob(os.path.join("kernels", "*.yml"))
    compile_kernels(kernel_specs)



# -*- coding: utf-8 -*-

from __future__ import division, print_function, absolute_import

__all__ = [
    "test_pickle", "test_pickle",
]

import sys
import pytest
import pickle
import numpy as np

from george import GP, kernels, BasicSolver, HODLRSolver


def _fake_compute(arg, *args, **kwargs):
    assert 0, "Unpickled GP shouldn't need to be computed"


@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3")
@pytest.mark.parametrize("solver,success", [(BasicSolver, True),
                                            (HODLRSolver, False)])
def test_pickle(solver, success, N=50, seed=123):
    np.random.seed(seed)
    kernel = 0.1 * kernels.ExpSquaredKernel(1.5)
    kernel.pars = [1, 2]
    gp = GP(kernel, solver=solver)
    x = np.random.rand(100)
    gp.compute(x, 1e-2)

    s = pickle.dumps(gp, -1)
    gp = pickle.loads(s)
    if success:
        gp.compute = _fake_compute
    gp.lnlikelihood(np.sin(x))



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["test_trivial_solver", "test_basic_solver", "test_hodlr_solver"]

import numpy as np

import george
from george.utils import nd_sort_samples
from george import kernels
from george import TrivialSolver, BasicSolver, HODLRSolver


def test_trivial_solver(N=300, seed=1234):
    # Sample some data.
    np.random.seed(seed)
    x = np.random.randn(N, 3)
    yerr = 1e-3 * np.ones(N)
    y = np.sin(np.sum(x, axis=1))

    solver = TrivialSolver()
    solver.compute(x, yerr)

    assert np.allclose(solver.log_determinant, 2*np.sum(np.log(yerr)))
    assert np.allclose(solver.apply_inverse(y), y / yerr**2)


def _test_solver(Solver, N=300, seed=1234, **kwargs):
    # Set up the solver.
    kernel = 1.0 * kernels.ExpSquaredKernel(1.0)
    solver = Solver(kernel, **kwargs)

    # Sample some data.
    np.random.seed(seed)
    x = np.atleast_2d(np.sort(10*np.random.randn(N))).T
    yerr = np.ones(N)
    solver.compute(x, yerr)

    # Build the matrix.
    K = kernel.get_value(x)
    K[np.diag_indices_from(K)] += yerr ** 2

    # Check the determinant.
    sgn, lndet = np.linalg.slogdet(K)
    assert sgn == 1.0, "Invalid determinant"
    assert np.allclose(solver.log_determinant, lndet), "Incorrect determinant"

    y = np.sin(x[:, 0])
    b0 = np.linalg.solve(K, y)
    b = solver.apply_inverse(y).flatten()
    assert np.allclose(b, b0)

    # Check the inverse.
    assert np.allclose(solver.apply_inverse(K), np.eye(N)), "Incorrect inverse"

def test_basic_solver(**kwargs):
    _test_solver(BasicSolver, **kwargs)


def test_hodlr_solver(**kwargs):
    _test_solver(HODLRSolver, tol=1e-10, **kwargs)

def test_strange_hodlr_bug():
    np.random.seed(1234)
    x = np.sort(np.random.uniform(0, 10, 50000))
    yerr = 0.1 * np.ones_like(x)
    y = np.sin(x)

    kernel = np.var(y) * kernels.ExpSquaredKernel(1.0)

    gp_hodlr = george.GP(kernel, solver=HODLRSolver, seed=42)
    n = 200
    gp_hodlr.compute(x[:n], yerr[:n])
    gp_hodlr.log_likelihood(y[:n])



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["Metric", "Subspace"]

import numpy as np
from scipy.linalg import cho_factor

from .modeling import Model


class Subspace(object):

    def __init__(self, ndim, axes=None):
        self.ndim = int(ndim)
        if axes is None:
            axes = np.arange(self.ndim)
        self.axes = np.atleast_1d(axes).astype(int)
        if np.any(self.axes >= self.ndim):
            raise ValueError("invalid axis for {0} dimensional metric"
                             .format(self.ndim))


class Metric(Model):

    def __init__(self,
                 metric,
                 bounds=None,
                 ndim=None,
                 axes=None,
                 lower=True):
        if isinstance(metric, Metric):
            self.metric_type = metric.metric_type
            self.parameter_names = metric.parameter_names
            self.unfrozen_mask = metric.unfrozen_mask
            self.set_parameter_vector(
                metric.get_parameter_vector(include_frozen=True),
                include_frozen=True)
            self.parameter_bounds = metric.parameter_bounds
            self.ndim = metric.ndim
            self.axes = metric.axes
            return

        if ndim is None:
            raise ValueError("missing required parameter 'ndim'")

        # Conform with the modeling protocol.
        parameter_names = []
        parameters = []

        # Save the number of dimensions.
        subspace = Subspace(ndim, axes=axes)
        self.ndim = subspace.ndim
        self.axes = subspace.axes

        # See if the parameter is a scalar.
        try:
            metric = float(metric)

        except TypeError:
            metric = np.atleast_1d(metric)

            # If the metric is a vector, it is meant to be axis aligned.
            if len(metric.shape) == 1:
                self.metric_type = 1
                if len(metric) != len(self.axes):
                    raise ValueError("dimension mismatch")
                if np.any(metric <= 0.0):
                    raise ValueError("invalid (negative) metric")
                for i, v in enumerate(metric):
                    parameter_names.append("log_M_{0}_{0}".format(i))
                    parameters.append(np.log(v))
            elif len(metric.shape) == 2:
                self.metric_type = 2
                if metric.shape[0] != metric.shape[1]:
                    raise ValueError("metric must be square")
                if len(metric) != len(self.axes):
                    raise ValueError("dimension mismatch")

                # Compute the Cholesky factorization and log the diagonal.
                params = cho_factor(metric, lower=True)[0]
                i = np.diag_indices_from(params)
                params[i] = np.log(params[i])
                params = params[np.tril_indices_from(params)]

                # Save the parameter vector.
                k = 0
                for i in range(len(self.axes)):
                    parameter_names.append("log_L_{0}_{0}".format(i))
                    parameters.append(params[k])
                    k += 1
                    for j in range(i+1, len(self.axes)):
                        parameter_names.append("L_{0}_{1}".format(i, j))
                        parameters.append(params[k])
                        k += 1
            else:
                raise ValueError("invalid metric dimensions")

        else:
            self.metric_type = 0
            parameter_names.append("log_M_0_0")
            parameters.append(np.log(metric))

        self.parameter_names = tuple(parameter_names)
        kwargs = dict(zip(parameter_names, parameters))
        if bounds is not None:
            kwargs["bounds"] = bounds
        super(Metric, self).__init__(**kwargs)

    def to_matrix(self):
        vector = self.get_parameter_vector(include_frozen=True)
        if self.metric_type == 0:
            return np.exp(vector) * np.eye(len(self.axes))
        elif self.metric_type == 1:
            return np.diag(np.exp(vector))
        else:
            n = len(self.axes)
            L = np.zeros((n, n))
            L[np.tril_indices_from(L)] = vector
            i = np.diag_indices_from(L)
            L[i] = np.exp(L[i])
            return np.dot(L, L.T)

    def __repr__(self):
        vector = self.get_parameter_vector(include_frozen=True)
        if self.metric_type == 0:
            params = ["{0}".format(float(np.exp(vector)))]
        elif self.metric_type == 1:
            params = ["{0}".format(repr(np.exp(vector)))]
        elif self.metric_type == 2:
            params = ["{0}".format(repr(self.to_matrix().tolist()))]
        params += ["ndim={0}".format(self.ndim),
                   "axes={0}".format(repr(self.axes))]
        params += ["bounds={0}".format([
            (None if a is None else np.exp(a),
             None if b is None else np.exp(b))
            for a, b in self.get_parameter_bounds(include_frozen=True)
        ])]
        return "Metric({0})".format(", ".join(params))



# -*- coding: utf-8 -*-

__all__ = [
    "__version__",
    "kernels",
    "GP",
    "Metric",
    "TrivialSolver",
    "BasicSolver",
    "HODLRSolver",
]

from .george_version import version as __version__

from . import kernels
from .gp import GP
from .metrics import Metric
from .solvers import TrivialSolver, BasicSolver, HODLRSolver



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = [
    "test_constant_mean",
    "test_callable_mean",
    "test_gp_mean",
    "test_gp_white_noise",
    "test_gp_callable_mean",
    "test_parameters",
    "test_bounds",
]

import pytest
import numpy as np

from george import GP, kernels
from george.utils import check_gradient
from george.modeling import Model, ConstantModel, CallableModel


def test_constant_mean():
    m = ConstantModel(5.0)
    check_gradient(m, np.zeros(4))


def test_callable_mean():
    m = CallableModel(lambda x: 5.0 * x)
    check_gradient(m, np.zeros(4))


def test_gp_mean(N=50, seed=1234):
    np.random.seed(seed)
    x = np.random.uniform(0, 5)
    y = 5 + np.sin(x)
    gp = GP(10. * kernels.ExpSquaredKernel(1.3),
            mean=5.0, fit_mean=True)
    gp.compute(x)
    check_gradient(gp, y)


def test_gp_callable_mean(N=50, seed=1234):
    np.random.seed(seed)
    x = np.random.uniform(0, 5)
    y = 5 + np.sin(x)
    mean = CallableModel(lambda x: 5.0*x)
    gp = GP(10. * kernels.ExpSquaredKernel(1.3), mean=mean)
    gp.compute(x)
    check_gradient(gp, y)


def test_gp_white_noise(N=50, seed=1234):
    np.random.seed(seed)
    x = np.random.uniform(0, 5)
    y = 5 + np.sin(x)
    gp = GP(10. * kernels.ExpSquaredKernel(1.3),
            mean=5.0, fit_mean=True,
            white_noise=0.1, fit_white_noise=True)
    gp.compute(x)
    check_gradient(gp, y)


class LinearWhiteNoise(Model):
    parameter_names = ("m", "b")

    def get_value(self, x):
        return self.m * x + self.b

    @Model.parameter_sort
    def compute_gradient(self, x):
        return dict(m=x, b=np.ones(len(x)))


def test_gp_callable_white_noise(N=50, seed=1234):
    np.random.seed(seed)
    x = np.random.uniform(0, 5)
    y = 5 + np.sin(x)
    gp = GP(10. * kernels.ExpSquaredKernel(1.3), mean=5.0,
            white_noise=LinearWhiteNoise(-6, 0.01),
            fit_white_noise=True)
    gp.compute(x)
    check_gradient(gp, y)

    gp.freeze_parameter("white_noise:m")
    check_gradient(gp, y)


def test_parameters():
    kernel = 10 * kernels.ExpSquaredKernel(1.0)
    kernel += 0.5 * kernels.RationalQuadraticKernel(log_alpha=0.1, metric=5.0)
    gp = GP(kernel, white_noise=LinearWhiteNoise(1.0, 0.1))

    n = len(gp.get_parameter_vector())
    assert n == len(gp.get_parameter_names())
    assert n - 2 == len(kernel.get_parameter_names())

    gp.freeze_parameter(gp.get_parameter_names()[0])
    assert n - 1 == len(gp.get_parameter_names())
    assert n - 1 == len(gp.get_parameter_vector())

    gp.freeze_all_parameters()
    assert len(gp.get_parameter_names()) == 0
    assert len(gp.get_parameter_vector()) == 0

    gp.kernel.thaw_all_parameters()
    gp.white_noise.thaw_all_parameters()
    assert n == len(gp.get_parameter_vector())
    assert n == len(gp.get_parameter_names())

    assert np.allclose(kernel[0], np.log(10.))


def test_bounds():
    kernel = 10 * kernels.ExpSquaredKernel(1.0, metric_bounds=[(None, 4.0)])
    kernel += 0.5 * kernels.RationalQuadraticKernel(log_alpha=0.1, metric=5.0)
    gp = GP(kernel, white_noise=LinearWhiteNoise(1.0, 0.1))

    # Test bounds length.
    assert len(gp.get_parameter_bounds()) == len(gp.get_parameter_vector())
    gp.freeze_all_parameters()
    gp.thaw_parameter("white_noise:m")
    assert len(gp.get_parameter_bounds()) == len(gp.get_parameter_vector())

    # Test invalid bounds specification.
    with pytest.raises(ValueError):
        kernels.ExpSine2Kernel(gamma=0.1, log_period=5.0, bounds=[10.0])



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = [
]

import pytest
import numpy as np

from george import kernels, GP
from george.solvers import hodlr

@pytest.mark.xfail
def test_custom_cholesky(seed=1234, ndim=5):
    np.random.seed(seed)

    # Build the matrix.
    L1 = np.random.randn(ndim, ndim)
    L1[np.diag_indices(ndim)] = np.exp(L1[np.diag_indices(ndim)])
    L1[np.triu_indices(ndim, 1)] = 0.0

    for L in (L1, np.eye(ndim)):
        A = np.dot(L, L.T)
        b = np.random.randn(ndim)

        Lvec = np.array(L)
        Lvec[np.diag_indices(ndim)] = 1. / Lvec[np.diag_indices(ndim)]

        Lb1 = np.linalg.solve(L, b)
        Lb2 = hodlr.custom_forward_sub(Lvec[np.tril_indices(ndim)],
                                       np.array(b))
        assert np.allclose(Lb1, Lb2)

        Ainvb1 = np.linalg.solve(A, b)
        Ainvb2 = hodlr.custom_backward_sub(Lvec[np.tril_indices(ndim)], Lb2)
        assert np.allclose(Ainvb1, Ainvb2)


def _general_metric(metric, N=100, ndim=3):
    kernel = 0.1 * kernels.ExpSquaredKernel(metric, ndim=ndim)

    x = np.random.rand(N, ndim)
    M0 = kernel.get_value(x)

    gp = GP(kernel)
    M1 = gp.get_matrix(x)
    assert np.allclose(M0, M1)

    # Compute the expected matrix.
    M2 = np.empty((N, N))
    for i in range(N):
        for j in range(N):
            r = x[i] - x[j]
            r2 = np.dot(r, np.linalg.solve(metric, r))
            M2[i, j] = 0.1 * np.exp(-0.5*r2)

    if not np.allclose(M0, M2):
        print(M0)
        print()
        print(M2)
        print()
        print(M0 - M2)
        print()
        print(M0 / M2)

        L = np.linalg.cholesky(metric)
        i = N - 1
        j = N - 2
        r = x[j] - x[i]
        print(x[i], x[j])
        print("r = ", r)
        print("L.r = ", np.dot(L, r))
    assert np.allclose(M0, M2)


def test_general_metric(seed=1234, N=2, ndim=3):
    np.random.seed(seed)

    _general_metric(np.eye(ndim), N=N, ndim=ndim)

    L = np.random.randn(ndim, ndim)
    L[np.diag_indices(ndim)] = np.exp(L[np.diag_indices(ndim)])
    L[np.triu_indices(ndim, 1)] = 0.0
    metric = np.dot(L, L.T)
    _general_metric(metric, N=N, ndim=ndim)


def test_axis_algined_metric(seed=1234, N=100, ndim=3):
    np.random.seed(seed)

    kernel = 0.1 * kernels.ExpSquaredKernel(np.ones(ndim), ndim=ndim)

    x = np.random.rand(N, ndim)
    M0 = kernel.get_value(x)

    gp = GP(kernel)
    M1 = gp.get_matrix(x)
    assert np.allclose(M0, M1)

    # Compute the expected matrix.
    M2 = 0.1*np.exp(-0.5*np.sum((x[None, :, :] - x[:, None, :])**2, axis=-1))
    assert np.allclose(M0, M2)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

import pytest
import numpy as np

from george import kernels, GP


def test_dtype(seed=123):
    np.random.seed(seed)
    kernel = 0.1 * kernels.ExpSquaredKernel(1.5)
    kernel.pars = [1, 2]
    gp = GP(kernel)
    x = np.random.rand(100)
    gp.compute(x, 1e-2)

kernels_to_test = [
    kernels.ConstantKernel(log_constant=0.1),
    kernels.ConstantKernel(log_constant=10.0, ndim=2),
    kernels.ConstantKernel(log_constant=5.0, ndim=5),

    kernels.DotProductKernel(),
    kernels.DotProductKernel(ndim=2),
    kernels.DotProductKernel(ndim=5, axes=0),

    kernels.CosineKernel(log_period=1.0),
    kernels.CosineKernel(log_period=0.5, ndim=2),
    kernels.CosineKernel(log_period=0.5, ndim=2, axes=1),
    kernels.CosineKernel(log_period=0.75, ndim=5, axes=[2, 3]),

    kernels.ExpSine2Kernel(gamma=0.4, log_period=1.0),
    kernels.ExpSine2Kernel(gamma=12., log_period=0.5, ndim=2),
    kernels.ExpSine2Kernel(gamma=17., log_period=0.5, ndim=2, axes=1),
    kernels.ExpSine2Kernel(gamma=13.7, log_period=-0.75, ndim=5, axes=[2, 3]),
    kernels.ExpSine2Kernel(gamma=-0.7, log_period=0.75, ndim=5, axes=[2, 3]),
    kernels.ExpSine2Kernel(gamma=-10, log_period=0.75),

    kernels.LocalGaussianKernel(log_width=0.5, location=1.0),
    kernels.LocalGaussianKernel(log_width=0.1, location=0.5, ndim=2),
    kernels.LocalGaussianKernel(log_width=1.5, location=-0.5, ndim=2, axes=1),
    kernels.LocalGaussianKernel(log_width=2.0, location=0.75, ndim=5,
                                axes=[2, 3]),

    kernels.LinearKernel(order=0, log_gamma2=0.0),
    kernels.LinearKernel(order=2, log_gamma2=0.0),
    kernels.LinearKernel(order=2, log_gamma2=0.0),
    kernels.LinearKernel(order=5, log_gamma2=1.0, ndim=2),
    kernels.LinearKernel(order=3, log_gamma2=-1.0, ndim=5, axes=2),
    kernels.LinearKernel(order=0, log_gamma2=0.0) +
    kernels.LinearKernel(order=1, log_gamma2=-1.0) +
    kernels.LinearKernel(order=2, log_gamma2=-2.0),

    kernels.PolynomialKernel(order=0, log_sigma2=-10.0),
    kernels.PolynomialKernel(order=2, log_sigma2=-10.0),
    kernels.PolynomialKernel(order=2, log_sigma2=0.0),
    kernels.PolynomialKernel(order=5, log_sigma2=1.0, ndim=2),
    kernels.PolynomialKernel(order=3, log_sigma2=-1.0, ndim=5, axes=2),

    12. * kernels.ExpSine2Kernel(gamma=0.4, log_period=1.0, ndim=5),
    12. * kernels.ExpSquaredKernel(0.4, ndim=3) + 0.1,
]

@pytest.mark.parametrize("kernel", kernels_to_test)
def test_kernel(kernel, N=20, seed=123, eps=1.32e-6):
    np.random.seed(seed)
    t1 = np.random.randn(N, kernel.ndim)
    kernel.test_gradient(t1, eps=eps)
    kernel.test_gradient(t1, t1[:1], eps=eps)


@pytest.mark.parametrize("kernel", kernels_to_test)
def test_x_gradient_kernel(kernel, N=20, seed=123, eps=1.32e-6):
    np.random.seed(seed)
    t1 = np.random.randn(N, kernel.ndim)
    kernel.test_x1_gradient(t1, eps=eps)
    kernel.test_x1_gradient(t1, np.array(t1[:1]), eps=eps)
    kernel.test_x2_gradient(t1, eps=eps)
    kernel.test_x2_gradient(np.array(t1[:1]), t1, eps=eps)


stationary_kernels = [
    (kernels.ExpKernel, {}),
    (kernels.ExpSquaredKernel, {}),
    (kernels.Matern32Kernel, {}),
    (kernels.Matern52Kernel, {}),
    (kernels.RationalQuadraticKernel, dict(log_alpha=np.log(1.0))),
    (kernels.RationalQuadraticKernel, dict(log_alpha=np.log(0.1))),
    (kernels.RationalQuadraticKernel, dict(log_alpha=np.log(10.0))),
]

@pytest.mark.parametrize("kernel_type,kwargs", stationary_kernels)
def test_stationary(kernel_type, kwargs):
    def build_kernel(metric, **more):
        kws = dict(kwargs, **more)
        return kernel_type(metric=metric, **kws)

    kernel = build_kernel(0.1)
    test_kernel(kernel)
    test_x_gradient_kernel(kernel)

    kernel = build_kernel(1.0)
    test_kernel(kernel)
    test_x_gradient_kernel(kernel)

    kernel = build_kernel(10.0)
    test_kernel(kernel)
    test_x_gradient_kernel(kernel)

    kernel = build_kernel([1.0, 0.1, 10.0], ndim=3)
    test_kernel(kernel)
    test_x_gradient_kernel(kernel)

    kernel = build_kernel(1.0, ndim=3)
    test_kernel(kernel)
    test_x_gradient_kernel(kernel)

    with pytest.raises(ValueError):
        kernel = build_kernel([1.0, 0.1, 10.0, 500], ndim=3)

    kernel = build_kernel(1.0, ndim=3, axes=2)
    test_kernel(kernel)
    test_x_gradient_kernel(kernel)

    kernel = build_kernel(1.0, ndim=3, axes=2, block=(-0.1, 0.1))
    test_kernel(kernel)
    test_x_gradient_kernel(kernel)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = [
    "test_gradient", "test_prediction", "test_repeated_prediction_cache",
    "test_apply_inverse",
]

import pytest
import numpy as np
from itertools import product

from george import kernels, GP, BasicSolver, HODLRSolver

@pytest.mark.parametrize("solver,white_noise",
                         product([BasicSolver, HODLRSolver], [None, 0.1]))
def test_gradient(solver, white_noise, seed=123, N=305, ndim=3, eps=1.32e-3):
    np.random.seed(seed)

    # Set up the solver.
    kernel = 1.0 * kernels.ExpSquaredKernel(0.5, ndim=ndim)
    kwargs = dict()
    if white_noise is not None:
        kwargs = dict(white_noise=white_noise, fit_white_noise=True)
    if solver == HODLRSolver:
        kwargs["tol"] = 1e-8
    gp = GP(kernel, solver=solver, **kwargs)

    # Sample some data.
    x = np.random.rand(N, ndim)
    x = x[np.argsort(x[:, 0])]
    y = gp.sample(x)
    gp.compute(x, yerr=0.1)

    # Compute the initial gradient.
    grad0 = gp.grad_log_likelihood(y)
    vector = gp.get_parameter_vector()

    for i, v in enumerate(vector):
        # Compute the centered finite difference approximation to the gradient.
        vector[i] = v + eps
        gp.set_parameter_vector(vector)
        lp = gp.lnlikelihood(y)

        vector[i] = v - eps
        gp.set_parameter_vector(vector)
        lm = gp.lnlikelihood(y)

        vector[i] = v
        gp.set_parameter_vector(vector)

        grad = 0.5 * (lp - lm) / eps
        assert np.abs(grad - grad0[i]) < 5 * eps, \
            "Gradient computation failed in dimension {0} ({1})\n{2}" \
            .format(i, solver.__name__, np.abs(grad - grad0[i]))


@pytest.mark.parametrize("solver", [BasicSolver, HODLRSolver])
def test_prediction(solver, seed=42):
    """Basic sanity checks for GP regression."""

    np.random.seed(seed)

    kernel = kernels.ExpSquaredKernel(1.0)
    kwargs = dict()
    if solver == HODLRSolver:
        kwargs["tol"] = 1e-8
    gp = GP(kernel, solver=solver, white_noise=0.0, **kwargs)

    x0 = np.linspace(-10, 10, 500)
    x = np.sort(np.random.uniform(-10, 10, 300))
    gp.compute(x)

    y = np.sin(x)
    mu, cov = gp.predict(y, x0)

    Kstar = gp.get_matrix(x0, x)
    K = gp.get_matrix(x)
    K[np.diag_indices_from(K)] += 1.0
    mu0 = np.dot(Kstar, np.linalg.solve(K, y))
    print(np.abs(mu - mu0).max())
    assert np.allclose(mu, mu0)


def test_repeated_prediction_cache():
    kernel = kernels.ExpSquaredKernel(1.0)
    gp = GP(kernel)

    x = np.array((-1, 0, 1))
    gp.compute(x)

    t = np.array((-.5, .3, 1.2))

    y = x/x.std()
    mu0, mu1 = (gp.predict(y, t, return_cov=False) for _ in range(2))
    assert np.array_equal(mu0, mu1), \
        "Identical training data must give identical predictions " \
        "(problem with GP cache)."

    y2 = 2*y
    mu2 = gp.predict(y2, t, return_cov=False)
    assert not np.array_equal(mu0, mu2), \
        "Different training data must give different predictions " \
        "(problem with GP cache)."

    a0 = gp._alpha
    gp.kernel[0] += 0.1
    gp.recompute()
    gp._compute_alpha(y2, True)
    a1 = gp._alpha
    assert not np.allclose(a0, a1), \
        "Different kernel parameters must give different alphas " \
        "(problem with GP cache)."

    mu, cov = gp.predict(y2, t)
    _, var = gp.predict(y2, t, return_var=True)
    assert np.allclose(np.diag(cov), var), \
        "The predictive variance must be equal to the diagonal of the " \
        "predictive covariance."


@pytest.mark.parametrize("solver", [BasicSolver, HODLRSolver])
def test_apply_inverse(solver, seed=1234, N=201, yerr=0.1):
    np.random.seed(seed)

    # Set up the solver.
    kernel = 1.0 * kernels.ExpSquaredKernel(0.5)
    kwargs = dict()
    if solver == HODLRSolver:
        kwargs["tol"] = 1e-10
    gp = GP(kernel, solver=solver, **kwargs)

    # Sample some data.
    x = np.sort(np.random.rand(N))
    y = gp.sample(x)
    gp.compute(x, yerr=yerr)

    K = gp.get_matrix(x)
    K[np.diag_indices_from(K)] += yerr**2

    b1 = np.linalg.solve(K, y)
    b2 = gp.apply_inverse(y)
    assert np.allclose(b1, b2)

    y = gp.sample(x, size=5).T
    b1 = np.linalg.solve(K, y)
    b2 = gp.apply_inverse(y)
    assert np.allclose(b1, b2)


@pytest.mark.parametrize("solver", [BasicSolver, HODLRSolver])
def test_predict_single(solver, seed=1234, N=201, yerr=0.1):
    np.random.seed(seed)

    # Set up the solver.
    kernel = 1.0 * kernels.ExpSquaredKernel(0.5)
    kwargs = dict()
    if solver == HODLRSolver:
        kwargs["tol"] = 1e-8
    gp = GP(kernel, solver=solver, **kwargs)

    x = np.sort(np.random.rand(N))
    y = gp.sample(x)
    gp.compute(x, yerr=yerr)
    mu0, var0 = gp.predict(y, [0.0], return_var=True)
    mu, var = gp.predict(y, [0.0, 1.0], return_var=True)
    _, cov = gp.predict(y, [0.0, 1.0])
    assert np.allclose(mu0, mu[0])
    assert np.allclose(var0, var[0])
    assert np.allclose(var0, cov[0, 0])



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["TrivialSolver"]

import numpy as np
from ..kernels import EmptyKernel


class TrivialSolver(object):

    def __init__(self, kernel=None):
        if (kernel is not None and
                kernel.kernel_type != EmptyKernel.kernel_type):
            raise ValueError("the trivial solver doesn't work with a kernel")
        self.computed = False
        self.log_determinant = None

    def compute(self, x, yerr):
        self._ivar = 1.0 / yerr ** 2
        self.log_determinant = 2 * np.sum(np.log(yerr))
        self.computed = True

    def apply_inverse(self, y, in_place=False):
        if not in_place:
            y = np.array(y)
        y[:] *= self._ivar
        return y

    def dot_solve(self, y):
        return np.sum(y**2 * self._ivar)

    def apply_sqrt(self, r):
        return r * np.sqrt(self._ivar)
